#include "cc.h"
#include "ctk.h"
#include "ctk-draw.h"
#include "ctk-vncserver.h"
#include "ek.h"
#include "program-handler.h"

#include "uip-event.h"
#include "uip.h"
#include "uip-fw.h"
#include "uip-fw-service.h"
#include "uip_arp.h"
#include "resolv.h"
#include "slipdev.h"
#include "rs232dev.h"
#include "webserver.h"
#include "about-dsc.h"
#include "netconf-dsc.h"
#include "processes-dsc.h"
#include "calc-dsc.h"
#include "www-dsc.h"
#include "webserver-dsc.h"
#include "telnet-dsc.h"
#include "vnc-dsc.h"

#include "ctk-x.h"
#include "clock.h"

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>

#define X_FRAMERATE     30000  // execute every 30ms
#define TIMER_FRAMERATE 1000   // execute every 1ms
#define RS232_FRAMERATE 5000   // execute every 5ms

static const struct uip_eth_addr ethaddr = {{0x00,0x06,0x98,0x01,0x02,0x26}};

static struct uip_fw_netif slipif =
  {UIP_FW_NETIF(0,0,0,0, 0,0,0,0, slipdev_send)};

static u16_t addr[2];

unsigned short thread_count = 0;
static pthread_mutex_t mr = PTHREAD_MUTEX_INITIALIZER;

clock_time_t
clock_time(void)
{
  u16_t tmp;
  pthread_mutex_lock(&mr);
  tmp = thread_count;
  pthread_mutex_unlock(&mr);
  return tmp;
}

void *thread_timer(void *t) 
{ 
   long time_padding;
   struct timeval now, then;

   while (1)
   {
       gettimeofday(&then, NULL);

       pthread_mutex_lock(&mr);
       thread_count++;
       pthread_mutex_unlock(&mr);

       gettimeofday(&now, NULL);
       time_padding = TIMER_FRAMERATE - ((now.tv_sec - then.tv_sec) * 1000000 +
                                   (now.tv_usec - then.tv_usec));
       if (time_padding > 0)
           usleep(time_padding);
   }
} 

  
void *thread_contiki(void *t) 
{
    ek_init();
    
    ctk_init();

    uip_event_init();

    tcpip_init(NULL);

    uip_ipaddr(addr, 10,0,0,2); // contiki
    uip_sethostaddr(addr);

    uip_ipaddr(addr, 10,0,0,1); // destination address
    uip_setdraddr(addr);

    uip_ipaddr(addr, 255,255,255,0);
    uip_setnetmask(addr);

    resolv_init(NULL);

//    uip_ipaddr(addr, 10,0,0,1); // dns server
//    resolv_conf(addr);
//    dnsconf(addr);
/*
  if(uiplib_ipaddrconv(str, (unsigned char *)addr)) {
    resolv_conf(addr);
  }
*/


    uip_fw_init();
    uip_fw_default(&slipif);
    uip_fw_service_init(NULL);

    program_handler_init();
    program_handler_add(&about_dsc, "About", 1);
    program_handler_add(&webserver_dsc, "Web server", 1);
    program_handler_add(&www_dsc, "Web browser", 1);
    program_handler_add(&calc_dsc, "Calculator", 1);
    program_handler_add(&telnet_dsc, "Telnet", 1);
    program_handler_add(&processes_dsc, "Processes", 0);
    program_handler_add(&netconf_dsc,"Network Setup", 1);

//    should do not compile ctk...
//    program_handler_add(&vnc_dsc,"VNC", 1);

    webserver_init(NULL);
    
    while(1)
    {
      if (ek_run() == 0)
      {
          /* Safe CPU */
          usleep(5000);
      }

      if (uip_len != 0)
          tcpip_input();
    }

    return 0;
} 

void *thread_rs232dev(void *t) 
{
    u16_t len;
    long time_padding;
    struct timeval now, then;

    rs232dev_init();

    while (1)
    {
        gettimeofday(&then, NULL);

        if (uip_len == 0)
        {
            len = slipdev_poll();
            if (len != 0)
                uip_len = len;
        }

        gettimeofday(&now, NULL);
        time_padding = RS232_FRAMERATE - ((now.tv_sec - then.tv_sec) * 1000000 +
                                    (now.tv_usec - then.tv_usec));
        if (time_padding > 0)
            usleep(time_padding);
    }

    // TODO : tratar SIGNAL para finalizar threads... 
    rs232dev_close();
}

/*-----------------------------------------------------------------------------------*/
int
main(int argc, char **argv)
{
    int res;
    pthread_t a_thread_timer;
    pthread_t a_thread_contiki;
    pthread_t a_thread_rs232dev;
    long time_padding;
    struct timeval now, then;

    Xsetup_window();
    Xsetup();

    printf("init timer thread...\r\n");
    res = pthread_create(&a_thread_timer, NULL, thread_timer, (void*)"hi timer");
    if (res != 0) {
        perror("Thread creation failed");
        exit(EXIT_FAILURE);
    }

    printf("init contiki thread...\r\n");
    res = pthread_create(&a_thread_contiki, NULL, thread_contiki, (void*)"hi contiki");
    if (res != 0) {
        perror("Thread creation failed");
        exit(EXIT_FAILURE);
    }

    printf("init rs232dev thread...\r\n");
    res = pthread_create(&a_thread_rs232dev, NULL, thread_rs232dev, (void*)"hi rs232dev");
    if (res != 0) {
        perror("Thread creation failed");
        exit(EXIT_FAILURE);
    }

    printf("executing X-eventloop...\r\n");
    
    while (1)
    {
        gettimeofday(&then, NULL);

        // process X events
        Xeventproc();

        gettimeofday(&now, NULL);
        time_padding = X_FRAMERATE - ((now.tv_sec - then.tv_sec) * 1000000 +
                                    (now.tv_usec - then.tv_usec));
        if (time_padding > 0)
            usleep(time_padding);
    }

    return 1;
}
/*-----------------------------------------------------------------------------------*/

