/*

/////////////////////////////////////////////
Global list of changes:

module ctk-conio.c:
- placed some definitions.

module libconio.c:
- cputc() now calls XPutc()

module dispatcher.c:
- replaced the function clock() - which is already defined in time.h - by the function cclock()
- added an usleep() call into dispatcher_run() - safe CPU.

module uipopt.h:
- UIP_LLH_LEN changed to support SLIP
/////////////////////////////////////////////

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define XK_MISCELLANY
#include <X11/keysymdef.h>

#include "ctk-x.h"
#include "ctk-arch.h"

char xserver[512];
Display * xdisplay;
int xscreen, xblack, xwhite, xhas_color, xfh, xfw;
XFontStruct * xfont;
Window xwindow, xroot;
GC whitegc, blackgc, yougc;
Visual * temp_visual;
Colormap xcolormap;
unsigned short xmouse_x, xmouse_y, xmouse_button;

#define WIDTH  640
#define HEIGHT 480
#define YOU_COLOR "LightYellow"

//typedef char ctk_arch_key_t;


/*-----------------------------------------------------------------------------------*/
void
ctk_arch_draw_char(char c,
		   unsigned char xpos,
		   unsigned char ypos,
		   unsigned char reversedflag,
		   unsigned char color)
{
    Xputc(c, xpos, ypos, reversedflag, color);
}
/*-----------------------------------------------------------------------------------*/
void
ctk_mouse_init(void)
{
    xmouse_x = 0;
    xmouse_y = 0;
    xmouse_button = 0;
}
/*-----------------------------------------------------------------------------------*/
unsigned char
ctk_mouse_button(void)
{
    return xmouse_button;
}
/*-----------------------------------------------------------------------------------*/
unsigned short
ctk_mouse_x(void)
{
    return xmouse_x;
}
/*-----------------------------------------------------------------------------------*/
unsigned short
ctk_mouse_y(void)
{
    return xmouse_y;
}
/*-----------------------------------------------------------------------------------*/
unsigned char
ctk_mouse_xtoc(unsigned short x)
{
    return (x / xfw) - 1;
}
/*-----------------------------------------------------------------------------------*/
unsigned char
ctk_mouse_ytoc(unsigned short y)
{
    return y / (xfh - 2);
}
/*-----------------------------------------------------------------------------------*/
void
ctk_mouse_hide(void)
{
}
/*-----------------------------------------------------------------------------------*/
void
ctk_mouse_show(void)
{
}
/*-----------------------------------------------------------------------------------*/
#define NUMKEYS 100
static ctk_arch_key_t keys[NUMKEYS];
static int firstkey = 0, lastkey = 0;

unsigned char
ctk_arch_keyavail(void)
{
  return firstkey != lastkey;
}
/*-----------------------------------------------------------------------------------*/
ctk_arch_key_t
ctk_arch_getkey(void)
{
  ctk_arch_key_t key;
  key = keys[firstkey];

  if(firstkey != lastkey) {
    ++firstkey;
    if(firstkey >= NUMKEYS) {
      firstkey = 0;
    }
  }
  
  return key;  
}
/*-----------------------------------------------------------------------------------*/
void add_key(ctk_arch_key_t key)
{
    keys[lastkey] = key;
    ++lastkey;
    if(lastkey >= NUMKEYS) {
      lastkey = 0;
    }
}
/*-----------------------------------------------------------------------------------*/

void Xputc(char c,
           unsigned char x,
           unsigned char y,
           unsigned char reversed,
           unsigned char color)
{
    char str[2];

    str[0] = c;
    str[1] = 0;

    x++;
    y++;

    //printf("[%c] x:%u y:%u rev:%u color:%u\r\n",c,x,y,reversed,color);

    if (reversed)
    {
        XDrawImageString(xdisplay, xwindow, whitegc,
                    x * (xfw /*- 6*/),
                    y * (xfh - 2), str, 1);
    }
    else
    {
        XDrawImageString(xdisplay, xwindow, blackgc,
                    x * (xfw /*- 6*/),
                    y * (xfh - 2), str, 1);
    }
}

void Xeventproc(void)
{
    XEvent event;
    char buffer[20];
    int bufsize = 20;
    KeySym key;
    XComposeStatus compose;
    int charcount;

    while (XPending(xdisplay))
    {
        XNextEvent(xdisplay, &event);

        if (event.type == KeyPress)
        {
            charcount = XLookupString((XKeyEvent *)&event, buffer, bufsize, &key, &compose);
            
            printf("KeyPress: %4X - ", (unsigned int)key);

            if (key < 255)
            {
                printf("printable char\r\n");
                add_key(key);
            }
            else
            {
                switch (key)
                {
                case XK_BackSpace:
                    printf("backspace\r\n");
                    add_key(0x08);
                    break;
                case XK_Up:
                    printf("up\r\n");
                    add_key(CH_CURS_UP);
                    break;
                case XK_Down:
                    printf("down\r\n");
                    add_key(CH_CURS_DOWN);
                    break;
                case XK_Left:
                    printf("left\r\n");
                    add_key(CH_CURS_LEFT);
                    break;
                case XK_Right:
                    printf("right\r\n");
                    add_key(CH_CURS_RIGHT);
                    break;
                case XK_Return:
                    printf("return\r\n");
                    add_key(CH_ENTER);
                    break;
                case XK_Delete:
                    printf("delete\r\n");
                    add_key(CH_DEL);
                    break;
                case XK_Escape:
                    printf("escape\r\n");
                    add_key(CH_ESC);
                    break;
                case XK_Home:
                    printf("home\r\n");
                    add_key(CH_HOME);
                    break;
                case XK_Tab:
                    printf("tab\r\n");
                    add_key(CH_TAB);
                    break;
                default:
                    printf("key not defined\r\n");
                }
            }
        }
        else if (event.type == Expose)
        {
            printf("Expose...\r\n");
            ctk_desktop_redraw(NULL);
        }
        else if (event.type == ButtonPress)
        {
            xmouse_button = 1;
        }
        else if (event.type == ButtonRelease)
        {
            xmouse_button = 0;
        }
        else if (event.type == MotionNotify)
        {
            xmouse_x = event.xmotion.x;
            xmouse_y = event.xmotion.y;
        }

    }
}

void Xsetup(void)
{
    int temp_depth;

    /* Set up visual: */

    xhas_color = 0;

    if (SetUpVisual(xdisplay, xscreen, &temp_visual, &temp_depth))
    {
        if (!SetUpColormap(xdisplay, xscreen, xwindow,
                           temp_visual, &xcolormap))
        {
            fprintf(stderr, "Could not create a colormap!\n");
        }
        else
            xhas_color = 1;
    }
    else
    {
        fprintf(stderr, "Could not find a PseudoColor visual!\n");
    }
/*  
  yougc   = CreateGC(xdisplay, xwindow,
            MyAllocNamedColor(xdisplay, xcolormap,
                      YOU_COLOR, xwhite,
                      xhas_color),
            MyAllocNamedColor(xdisplay, xcolormap,
                      xobject_colors, xblack,
                      xhas_color));
*/  
    /* Put us up!: */

    XMapWindow(xdisplay, xwindow);
    XMapRaised(xdisplay, xwindow);
    XSync(xdisplay, 0);
}


unsigned long MyAllocNamedColor(Display *display, Colormap colormap,
                                char* colorname, unsigned long default_color,
                                int has_color)
{
    if (has_color == 0)
        return(default_color);
    else
        return(AllocNamedColor(display, colormap, colorname, default_color));
}


/* Connect, make window, and display "please wait" */

void Xsetup_window(void)
{
    char title[1024];
    int done;
    XEvent event;

    /* get the DISPLAY env */

    if (getenv("DISPLAY") != NULL)
        strcpy(xserver, getenv("DISPLAY"));
    else
    {
        fprintf(stderr, "Can't determine your display!\n");
        fprintf(stderr, "Please specify it on the command line.\n");
        exit(1);
    }

    /* Connect to display: */

    fprintf(stderr, "Connecting to: %s\n", xserver);

    xdisplay = ConnectToServer(xserver, &xscreen, &xroot);
    if (xdisplay == NULL)
    {
        perror(xserver);
        exit(1);
    }


    /* Load font: */
//  xfont = LoadFont(xdisplay, "variable", "fixed");
    xfont = LoadFont(xdisplay, "fixed", "variable");
    xfh = FontHeight(xfont);
    xfw = FontWidth(xfont);
    printf("Xfont width: %u height %u\r\n", xfw, xfh);


    /* Get our primitve colors: */

    xblack = BlackPixel(xdisplay, xscreen);
    xwhite = WhitePixel(xdisplay, xscreen);


    /* Open window: */

    xwindow = OpenWindow(xdisplay, xroot, 0, 0,
                         WIDTH, HEIGHT,
                         CopyFromParent, xblack,
                         (KeyPressMask | ExposureMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask),
                         (Visual *)CopyFromParent);

    sprintf(title, "Contiki-X");
    SetStandardHints(xdisplay, xwindow, "Contiki-X", title, 10, 10,
                     WIDTH, HEIGHT);


    /* Create Primitive GC's: */

    whitegc = CreateGC(xdisplay, xwindow,
                       xwhite, xblack);

    blackgc = CreateGC(xdisplay, xwindow,
                       xblack, xwhite);



    /* Put us up!: */

    XMapWindow(xdisplay, xwindow);
    XMapRaised(xdisplay, xwindow);
    XSync(xdisplay, 0);


    /* Wait for map event: */

    done = 0;

    while (done == 0)
    {
        if (XPending(xdisplay))
        {
            XNextEvent(xdisplay, &event);

            if (event.type == Expose || event.type == MapNotify)
            {
                done = 1;
            }
        }
    }

    drawcenteredtext(xdisplay, xwindow,
                     whitegc, 0, WIDTH, HEIGHT / 2,
                     "Contiki-X : Please Wait", xfont);

    XSync(xdisplay, 0);
}

Display *ConnectToServer(char* display_name, int *screen, Window *rootwindow)
{
    Display *display;

    /* connect & check for problems */
    display = XOpenDisplay(display_name);
    if (display != NULL)
    {
        /* get screen # */
        *screen = DefaultScreen(display);

        /* get screens root window id */
        *rootwindow = RootWindow(display,*screen);
    }

    return(display);
}

int FontHeight(XFontStruct *font_struct)
{
    return(font_struct->ascent + font_struct->descent);
}

int FontWidth(XFontStruct *font_struct)
{
    return(font_struct->max_bounds.rbearing - font_struct->min_bounds.lbearing );
}

XFontStruct *LoadFont(Display *display, char* font_name,
                      char* fallback_font_name)
{
    XFontStruct *font_struct;

    font_struct = XLoadQueryFont(display, font_name);
    if (font_struct == (XFontStruct*)NULL)
    {
        /* try to load the fallbacck font */
        font_struct = XLoadQueryFont(display,fallback_font_name);
        if (font_struct != (XFontStruct*)NULL)
            printf("Warning: cant load font (fallback font loaded...)\r\n");
    }

    return(font_struct);
}


void drawtext(Display *display, Window window, GC gc, int x, int y, char *s)
{
    XDrawImageString(display, window, gc, x, y, s, strlen(s));
}

void drawcenteredtext(Display *display, Window window, GC gc, int x1, int x2,
                      int y, char *s, XFontStruct *font)
{
    int x;

    x = ((x2 - x1) / 2) + x1;
    x = x - (XTextWidth(font, s, strlen(s))) / 2;

    drawtext(display, window, gc, x, y, s);
}

Window OpenWindow(Display *display, Window parent, int x, int y, int width,
                  int height, unsigned long bordercolor, 
                  unsigned long backcolor, unsigned long event_mask,
                  Visual *visual)
{
    Window window;
    XSetWindowAttributes attributes;
    unsigned long attr_mask;

    /* set up window attributes first */

    attributes.event_mask = event_mask;
    attributes.border_pixel = bordercolor;
    attributes.background_pixel = backcolor;
    attr_mask = CWEventMask | CWBackPixel | CWBorderPixel;

    /* uncomment this to override-redirect: */
    /*
      attributes.override_redirect = True;
      attr_mask = attr_mask | CWOverrideRedirect;
      */

    /* create window! */
    window = XCreateWindow(display,parent,x,y,width,height,BORDER_WIDTH,
                           DefaultDepthOfScreen(DefaultScreenOfDisplay(display)),
                           InputOutput,visual,attr_mask,
                           &attributes);

    return(window);
}

/* Local prototypes */

void SetClassHints(Display *display, Window window, char *res_name, 
                   char *res_class);
void FillSizeHints(int x, int y, int width, int height, XSizeHints *sizehints);
void SetWMHints(Display *display, Window window, int initial_state);

/* Local Definations */

void SetClassHints(Display *display, Window window, char *res_name,
                   char *res_class)
/*
   Sets a windows class hints.
*/
{
    XClassHint class_hints;

    class_hints.res_class = res_class;
    class_hints.res_name = res_name;

    XSetClassHint(display, window, &class_hints);
}

void FillSizeHints(int x, int y, int width, int height, XSizeHints *sizehints)
/*
   Fills in an XSizeHints struct
*/
{
    /* set size hints */
    sizehints->x = x;             /* obsolete */
    sizehints->y = y;             /* obsolete */
    sizehints->height = height;   /* obsolete */
    sizehints->width = width;     /* obsolete */
    sizehints->min_height = height;
    sizehints->min_width = width;
    sizehints->flags = USPosition | USSize | PMinSize;

    /* if R4, we can set base stuff, too... */
#ifdef X11R4
    sizehints->base_height = height;
    sizehints->base_width = width;
    sizehints->flags = sizehints->flags | PBaseSize;
#endif
}

void SetWMHints(Display *display, Window window, int initial_state)
/*
   Sets the window manager hints.  "initial_state" can be:
   Withdrawn (0), Normal (1) or Iconic (3).
*/
{
    XWMHints wm_hints;

    wm_hints.flags = InputHint | StateHint;
    wm_hints.initial_state = initial_state;
    wm_hints.input = True;

    XSetWMHints(display,window,&wm_hints);
}

/* Global Definations */

void SetSizeHints(Display *display, Window window, int x, int y, int width,
                  int height)
{
    XSizeHints sizehints;

    /* get hint values */
    FillSizeHints(x,y,width,height,&sizehints);

    /* test for R4 - adds XSetWMNormalHints */
#ifdef X11R4
    XSetWMNormalHints(display,window,&sizehints);
#else
    XSetNormalHints(display,window,&sizehints);
#endif
}

void SetStandardHints(Display *display, Window window, char* app_name,
                      char* wind_name, int x, int y, int width, int height)
{
    SetSizeHints(display,window,x,y,width,height);
    SetWindowName(display,window,wind_name);
    SetClassHints(display,window,app_name,APPL_CLASS);
    SetWMHints(display,window,NormalState);
}

void SetWindowName(Display *display, Window window, char* name)
{
    XStoreName(display, window, name);
}

GC CreateGC(Display *display, Drawable drawable, unsigned long forecolor,
            unsigned long backcolor)
{
    XGCValues xgcvalues;
    GC gc;

    xgcvalues.foreground = forecolor;
    xgcvalues.background = backcolor;
    gc = XCreateGC(display,drawable,(GCForeground | GCBackground),
                   &xgcvalues);

    return(gc);
}

GC CreateClearGC(Display *display, Drawable drawable)
{
    XGCValues xgcvalues;
    GC gc;

    xgcvalues.function = GXclear;
    xgcvalues.fill_style = FillSolid;
    gc = XCreateGC(display, drawable, (GCFunction), &xgcvalues);

    return(gc);
}

GC CreateXorGC(Display *display, Drawable drawable, unsigned long forecolor,
               unsigned long backcolor)
{
    XGCValues xgcvalues;
    GC gc;

    xgcvalues.foreground = forecolor ^ backcolor;
    xgcvalues.background = 0;
    xgcvalues.function = GXxor;

    gc = XCreateGC(display,drawable,(GCForeground | GCBackground | GCFunction),
                   &xgcvalues);

    return(gc);
}

GC CreateAndGC(Display *display, Drawable drawable, unsigned long forecolor,
               unsigned long backcolor)
{
    XGCValues xgcvalues;
    GC gc;

    xgcvalues.foreground = forecolor;
    xgcvalues.background = backcolor;
    xgcvalues.function = GXand;

    gc = XCreateGC(display,drawable,(GCForeground | GCBackground | GCFunction),
                   &xgcvalues);

    return(gc);
}


GC CreateAndInvertedGC(Display *display, Drawable drawable,
                       unsigned long forecolor, unsigned long backcolor)
{
    XGCValues xgcvalues;
    GC gc;

    xgcvalues.foreground = forecolor;
    xgcvalues.background = backcolor;
    xgcvalues.function = GXandInverted;

    gc = XCreateGC(display,drawable,(GCForeground | GCBackground | GCFunction),
                   &xgcvalues);

    return(gc);
}

GC CreateOrGC(Display *display, Drawable drawable, unsigned long forecolor,
              unsigned long backcolor)
{
    XGCValues xgcvalues;
    GC gc;

    xgcvalues.foreground = forecolor;
    xgcvalues.background = backcolor;
    xgcvalues.function = GXor;

    gc = XCreateGC(display,drawable,(GCForeground | GCBackground | GCFunction),
                   &xgcvalues);

    return(gc);
}

int SetUpVisual(Display *display, int screen, Visual **visual, int *depth)
{
    int number_visuals;
    XVisualInfo *visual_array;
    XVisualInfo visual_info_template;
    int status;

    status = False;

    /* Will the default visual work? */
    if (DefaultVisual(display,screen)->class == PseudoColor ||
        DefaultVisual(display,screen)->class == TrueColor)
    {
        *visual = DefaultVisual(display,screen);
        *depth = DefaultDepth(display,screen);
        status = True;
    }
    else
    {
        /* we look for a PseudoColor visual on the current screen */
        visual_info_template.class = PseudoColor;
        visual_info_template.screen = screen;
        visual_array = XGetVisualInfo(display,VisualClassMask | VisualScreenMask,
                                      &visual_info_template,&number_visuals);

        /* Did we find one? */
        if ((number_visuals>0) && (visual_array != NULL))
        {
            /* Choose the first PseudoColor visual (LAZY!) */
            *visual = visual_array[0].visual;
            *depth = visual_array[0].depth;
            XFree(visual_array);   /* done with this chunk of RAM */
            status = True;
        }
        else
        {
            *visual = CopyFromParent;
            status = False;  /* sorry - live with what's left */
        }
    }

    return(status);
}

unsigned long AllocNamedColor(Display *display, Colormap colormap,
                              char* colorname, unsigned long default_color)
{
    XColor hardwarecolor,exactcolor;
    unsigned long color;
    int status;

    status = XAllocNamedColor(display,colormap,colorname,&hardwarecolor,
                              &exactcolor);
    if (status != 0)
        color = hardwarecolor.pixel;
    else
        color = default_color;

    return(color);
}

int SetUpColormap(Display *display, int screen, Window window, Visual *visual,
                  Colormap *colormap)
{
    int status = False;

    if (visual == DefaultVisual(display,screen))
    {
        *colormap = DefaultColormap(display,screen);
        status = True;
    }
    else
    {
        *colormap = XCreateColormap(display,window,visual,AllocNone);
        if (*colormap != None)
        {
            XSetWindowColormap(display,window,*colormap);
            status = True;
        }
        else
            *colormap = DefaultColormap(display,screen);
    }

    return(status);
}

