#ifndef CTK_X_H
#define CTK_X_H

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include "ctk.h"

#define BORDER_WIDTH 2
#define APPL_CLASS "Examples"

void Xputc(char c,
			unsigned char xpos,
			unsigned char ypos,
			unsigned char reversedflag,
			unsigned char color);

void Xeventproc(void);
void Xsetup(void);
void Xsetup_window(void);
unsigned long MyAllocNamedColor(Display *display, Colormap colormap,
                                char* colorname, unsigned long default_color,
                                int has_color);

int FontHeight(XFontStruct *font_struct);
/*
   Returns height of text in a given font.
*/


int FontWidth(XFontStruct *font_struct);
/*
   Returns width of text in a given font.
*/

XFontStruct *LoadFont(Display *display, char* font_name,
		      char* fallback_font_name);
/*
   Loads an X font into an XFontStruct and returns a pointer to the 
   struct.  If "font_name" can't be loaded, it tries to load
   "fallback_font_name".  If THAT fails too, it will return NULL.
*/

void drawtext(Display *display, Window window, GC gc, int x, int y, char *s);

/*
   Draws string 's' into 'window' on a 'display' at an 'x,y' location using
   the graphic context 'gc'
*/

void drawcenteredtext(Display *display, Window window, GC gc, int x1, int x2,
		      int y, char *s, XFontStruct *font);
/*
  Draws text centered
  */

Display *ConnectToServer(char* display_name, int *screen, Window *rootwindow);
/* 
   Connects to an X server, gets default screen # and root window id for
   that display (server) and screen 
	
   returns: Pointer to the display or NULL if there was a problem.
*/

Window OpenWindow(Display *display, Window parent, int x, int y, int width,
		  int height, unsigned long bordercolor, 
		  unsigned long backcolor, unsigned long event_mask,
		  Visual *visual);
/*
   Creates a window.
	
   returns: Window id.
*/

void SetSizeHints(Display *display, Window window, int x, int y, int width,
		  int height);
/*
   Sends the size hints to the WM
*/

void SetStandardHints(Display *display, Window window, char* app_name,
		      char* wind_name, int x, int y, int width, int height);
/*
   Makes suggestions to the WM about the look of new windows.
*/ 

void SetWindowName(Display *display, Window window, char* name);
/*
   Sets a window's name.
*/

GC CreateGC(Display *display, Drawable drawable, unsigned long forecolor,
            unsigned long backcolor);
/* 
   creates a graphics context with the given foreground and background
   pixel values.
   
   Returns a GC
*/

GC CreateClearGC(Display *display, Drawable drawable);
/* 
   creates a graphics context with a clear function.
   
   Returns a GC
*/

GC CreateXorGC(Display *display, Drawable drawable, unsigned long forecolor,
	       unsigned long backcolor);
/*
   Creates an XOR graphics context.

   Returns a GC
*/

GC CreateAndGC(Display *display, Drawable drawable, unsigned long forecolor,
	       unsigned long backcolor);
/*
   Creates an AND graphics context.

   Returns a GC
*/

GC CreateAndInvertedGC(Display *display, Drawable drawable,
		       unsigned long forecolor, unsigned long backcolor);
/*
   Creates an AND_Inverted graphics context.

   Returns a GC
*/

GC CreateOrGC(Display *display, Drawable drawable, unsigned long forecolor,
	      unsigned long backcolor);
/*
   Creates an Or graphics context.

   Returns a GC
*/

int SetUpVisual(Display *display, int screen, Visual **visual, int *depth);
/*
   Grabs a PseudoColor visual.
*/

unsigned long AllocNamedColor(Display *display, Colormap colormap,
                              char* colorname, unsigned long default_color);
/*
   Allocates a read-only (sharable) color cell in the given colormap,
   using the RGB database to convert the color name to an RGB value.
   See /usr/lib/X11/rgb.txt for the database listing.
*/


int SetUpColormap(Display *display, int screen, Window window, Visual *visual,
		  Colormap *colormap);
/*
   Create an X colormap using the given screen and visual.  Uses the
   default colormap, if possible, to avoid palette switches when
   other windows are brought "forward".
*/

#endif
