#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>

#include "rs232dev.h"

typedef unsigned char u8_t;

/*-----------------------------------------------------------------------------------*/

int mytty = 0;

/*-----------------------------------------------------------------------------------*/

void tty_open (char* ttyport, speed_t baudrate,unsigned int timeout,
               unsigned int ignore_serial_bits);

/*-----------------------------------------------------------------------------------*/

u8_t slipdev_char_poll (u8_t *c)
{
    if (mytty == 0) return 0;

    if (read(mytty,c,1))
        return 1;
    else
        return 0;
}

/*-----------------------------------------------------------------------------------*/

void slipdev_char_put (u8_t c)
{
    if (mytty == 0) return;

    while (write(mytty,&c,1) != 1);
}

/*-----------------------------------------------------------------------------------*/

void rs232dev_init(void)
{
    int i = 0;
    char newdata;
    const char client_str[]="CLIENT";
    const char server_str[]="CLIENTSERVER";
    const char device_str[]="/dev/ttyS1"; // does not work with old cygwin DLL's   
                                          // you can try "COM2" or "/dev/COM2" ... 
    printf("\r\nOPEN %s, 38400\r\n", device_str);
    tty_open((char *)device_str, B9600, 1, 0);  
                                          
    printf("Waiting CLIENT string\r\n");

    /* Wait the "CLIENT" string */
    while (i < 6)
    {
        while(read(mytty,&newdata,1) == 0);
        printf("%c\r\n", newdata);

        if (client_str[i] == newdata)
            i++;
        else
            i = 0;
        if (i == 6) break;
    }

    /* Send the "CLIENTSERVER" string */
    printf("\r\nSending CLIENTSERVER string\r\n");

    i = 0;
    while (server_str[i] != 0)
    {
        write(mytty,&server_str[i],1);
        i++;
    }

    return;
}

/*-----------------------------------------------------------------------------------*/

void rs232dev_close (void)
{
    close(mytty);
}

/*-----------------------------------------------------------------------------------*/

int cfmakeraw(struct termios *termios_p) {
  termios_p->c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
  termios_p->c_oflag &= ~OPOST;
  termios_p->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
  termios_p->c_cflag &= ~(CSIZE|PARENB);
  termios_p->c_cflag |= CS8;
  return 1;
}

/*-----------------------------------------------------------------------------------*/

void tty_open(char* ttyport, speed_t baudrate, unsigned int timeout,
              unsigned int ignore_serial_bits){
  int flags=O_RDWR | O_NOCTTY | O_NONBLOCK;
  //  int flags=O_RDWR | O_NOCTTY;
  struct termios newtio;

  //opening the device
  printf("Accessing device %s\n",ttyport);
	
  if ((mytty=open(ttyport,flags)) == -1){
    printf("Error on opening device.\n%s\n",strerror(errno));
    exit(1);
  }

  if (fcntl(mytty,F_SETFL,0) == -1) {
    fprintf(stderr,"Error in setting port attributes: %s\n",strerror(errno));
    exit(1);
  }
  if (!ignore_serial_bits) {
    //getting port parameters
    if (tcgetattr(mytty,&newtio) < 0){
      fprintf(stderr,"Error in getting device attributes: %s\n",strerror(errno));
      exit(1);
    }
  } else {
    memset(&newtio,0,sizeof(newtio));
  }
  cfsetispeed(&newtio, baudrate); //input at baudrate
  cfsetospeed(&newtio, baudrate); //ouput at baudrate
  cfmakeraw(&newtio); //make raw, 8N1
  newtio.c_cflag |= (CRTSCTS | CREAD);    // for NULL MODEM cables replace "CRTSCTS" with "CLOCAL".
  newtio.c_cflag &= ~CSTOPB; // do not send two stop bits (but one)

  newtio.c_cc[VMIN]=0; //be comfortable with returning this amount of characters
  newtio.c_cc[VTIME]=timeout*10; //try reading for this amount of time (in deciseconds)
  tcflush(mytty, TCIOFLUSH); //clear serial tty
  if(tcsetattr(mytty,TCSANOW,&newtio) < 0){ //set now
    fprintf(stderr,"Error in setting device attributes: %s\n",strerror(errno));
    exit(1);
  }
}

/*-----------------------------------------------------------------------------------*/

